"""
Create a NeXus file for time-resolved collections on I19-2 using parameters passed from GDA.
"""

from __future__ import annotations

import logging
from pathlib import Path
from typing import Optional, Tuple

import numpy as np
from numpy.typing import DTypeLike

from .. import log
from ..nxs_utils import (
    Attenuator,
    Beam,
    Detector,
    EigerDetector,
    Goniometer,
    Source,
    TristanDetector,
)
from ..nxs_utils.detector import DetectorType, UnknownDetectorTypeError
from ..nxs_utils.scan_utils import calculate_scan_points
from ..nxs_write.nxmx_writer import EventNXmxFileWriter, NXmxFileWriter
from ..utils import find_in_dict, get_iso_timestamp, get_nexus_filename
from .beamline_utils import BeamlineAxes, GeneralParams, collection_summary_log
from .GDAtools.ExtendedRequest import (
    ExtendedRequestIO,
    read_det_position_from_xml,
    read_scan_from_xml,
)
from .GDAtools.GDAjson2params import JSONParamsIO

# Define a logger object and a formatter
logger = logging.getLogger("nexgen.I19-2_NeXus_gda")


class GDACollectionParams(GeneralParams):
    """Collection parameters for I19-2 with information extracted from GDA.

    Args:
        GeneralParams (Basemodel): General collection parameters common to \
            multiple beamlines/experiments, such as exposure time, wavelength, ...
        meta_file (Path | str): Path to _meta.h5 file.
        xml_file (Path | str):
        detector_name (str): Name of the detector in use for current experiment.
        geometry_json (Path | str, optional): Path to GDA config JSON file \
            describing the beamline geometry.
        detector_json (Path | str, optional): Path to GDA config JSON file \
            describing the detector.
    """

    meta_file: Path | str
    xml_file: Path | str
    detector_name: str
    geometry_json: Optional[Path | str]
    detector_json: Optional[Path | str]


def tristan_writer(
    master_file: Path,
    TR: GDACollectionParams,
    axes_params: BeamlineAxes,
    det_params: DetectorType,
    timestamps: Tuple[str, str] = (None, None),
):
    """
    A function to call the nexus writer for Tristan 10M detector.

    Args:
        master_file (Path): Path to nexus file to be written.
        TR (namedtuple): Parameters passed from the beamline.
        axes_params (BeamlineAxes): Axes for goniometer, detector and detector module.
        det_params (DetectorType): Detector definition for Tristan.
        timestamps (Tuple[str, str], optional): Collection start and end time. Defaults to None.
    """
    ecr = ExtendedRequestIO(TR.xml_file)
    # Read information from xml file
    logger.info("Read xml file.")
    scan_axis, pos, _ = read_scan_from_xml(ecr)
    # n_Frames is only useful for eiger
    # pos[scan_axis][::-1] is scan range
    scan_range = pos[scan_axis][:-1]
    # Define OSC scans dictionary
    OSC = {scan_axis: scan_range}

    det_positions = read_det_position_from_xml(ecr, det_params.description)

    # Attenuator
    attenuator = Attenuator(ecr.getTransmission())

    # Beam
    beam = Beam(TR.wavelength)

    # Source
    source = Source("I19-2")

    # Detector
    det_axes = axes_params.det_axes
    det_axes[0].start_pos = det_positions[0]  # two_theta
    det_axes[1].start_pos = det_positions[1]  # det_z
    detector = Detector(
        det_params,
        det_axes,
        TR.beam_center,
        TR.exposure_time,
        [axes_params.fast_axis, axes_params.slow_axis],
    )

    # Goniometer
    gonio_axes = axes_params.gonio
    for k, v in pos.items():
        # Get correct start positions
        idx = [n for n, ax in enumerate(gonio_axes) if ax.name == k][0]
        gonio_axes[idx].start_pos = v[0]
    goniometer = Goniometer(gonio_axes, OSC)

    collection_summary_log(
        logger,
        goniometer,
        detector,
        attenuator,
        beam,
        source,
        timestamps,
    )

    # Get on with the writing now...
    try:
        EventFileWriter = EventNXmxFileWriter(
            master_file,
            goniometer,
            detector,
            source,
            beam,
            attenuator,
        )
        EventFileWriter.write(start_time=timestamps[0])
        if timestamps[1]:
            EventFileWriter.update_timestamps(timestamps[1], "end_time")
        logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )


def eiger_writer(
    master_file: Path,
    TR: GDACollectionParams,
    axes_params: BeamlineAxes,
    det_params: DetectorType,
    timestamps: Tuple[str, str] = (None, None),
    vds_dtype: DTypeLike = np.uint16,
):
    """
    A function to call the nexus writer for Eiger 2X 4M detector.

    Args:
        master_file (Path): Path to nexus file to be written.
        TR (namedtuple): Parameters passed from the beamline.
        axes_params (BeamlineAxes): Axes for goniometer, detector and detector module.
        det_params (DetectorType): Detector definition for Eiger.
        timestamps (Tuple[str, str], optional): Collection start and end time. Defaults to (None, None).
        vds_dtype (DtypeLike): Data type for vds as np.uint##.
    """
    ecr = ExtendedRequestIO(TR.xml_file)
    # Read information from xml file
    logger.info("Read xml file.")
    scan_axis, pos, n_frames = read_scan_from_xml(ecr)

    det_positions = read_det_position_from_xml(ecr, det_params.description)

    # Attenuator
    attenuator = Attenuator(ecr.getTransmission())

    # Beam
    beam = Beam(TR.wavelength)

    # Source
    source = Source("I19-2")

    # Detector
    det_axes = axes_params.det_axes
    det_axes[0].start_pos = det_positions[0]  # two_theta
    det_axes[1].start_pos = det_positions[1]  # det_z
    detector = Detector(
        det_params,
        det_axes,
        TR.beam_center,
        TR.exposure_time,
        [axes_params.fast_axis, axes_params.slow_axis],
    )

    # Goniometer
    gonio_axes = axes_params.gonio
    for k, v in pos.items():
        # Get correct start positions
        idx = [n for n, ax in enumerate(gonio_axes) if ax.name == k][0]
        gonio_axes[idx].start_pos = v[0]

    # Get scan range array
    logger.info("Calculating scan range...")
    scan_idx = [n for n, ax in enumerate(gonio_axes) if ax.name == scan_axis][0]

    # Define OSC scans dictionary
    OSC = calculate_scan_points(
        gonio_axes[scan_idx], rotation=True, tot_num_imgs=n_frames
    )

    goniometer = Goniometer(gonio_axes, OSC)

    collection_summary_log(
        logger,
        gonio_axes,
        [scan_axis],
        detector,
        attenuator,
        beam,
        source,
        timestamps,
    )

    # Get on with the writing now...
    try:
        NXmx_Writer = NXmxFileWriter(
            master_file,
            goniometer,
            detector,
            source,
            beam,
            attenuator,
            n_frames,
        )
        NXmx_Writer.write(start_time=timestamps[0])
        if timestamps[1]:
            NXmx_Writer.update_timestamps(timestamps[1], "end_time")
        NXmx_Writer.write_vds(
            vds_shape=(n_frames, *detector.detector_params.image_size),
            vds_dtype=vds_dtype,
        )
        logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )


def write_nxs(**tr_params):
    """
    Gather all parameters from the beamline and call the NeXus writers.

    Keyword Args:
        meta_file (Path | str): Path to _meta.h5 file.
        xml_file (Path | str): Path to gda-generated xml file.
        detector_name (str): Detector in use.
        exposure_time (float): Exposure time, in s.
        wavelength (float): Wavelength of incident beam, in A.
        beam_center (List[float, float]): Beam center position, in pixels.
        start_time (datetime): Experiment start time.
        stop_time (datetime): Experiment end time.
        geometry_json (Path | str): Path to GDA generated geometry json file.
        detector_json (Path | str): Path to GDA generated detector json file.
        vds_dtype (DtypeLike): Data type for vds as np.uint##.
    """
    # Get info from the beamline
    TR = GDACollectionParams(
        meta_file=Path(tr_params["meta_file"]).expanduser().resolve(),
        xml_file=Path(tr_params["xml_file"]).expanduser().resolve(),
        detector_name=tr_params["detector_name"],
        exposure_time=tr_params["exposure_time"],
        wavelength=tr_params["wavelength"],
        beam_center=tr_params["beam_center"],
        geometry_json=(
            tr_params["geometry_json"]
            if find_in_dict(tr_params, "geometry_json")
            else None
        ),
        detector_json=(
            tr_params["detector_json"]
            if find_in_dict(tr_params, "detector_json")
            else None
        ),
    )

    # Define a file handler
    logfile = TR.meta_file.parent / "nexus_writer.log"
    # Configure logging
    log.config(logfile.as_posix())

    logger.info("NeXus file writer for beamline I19-2 at DLS.")
    logger.info(f"Detector in use for this experiment: {TR.detector_name}.")
    logger.info(f"Current collection directory: {TR.meta_file.parent}")
    # Add some information to logger
    logger.info("Creating a NeXus file for %s ..." % TR.meta_file.name)
    # Get NeXus filename
    master_file = get_nexus_filename(TR.meta_file)
    logger.info("NeXus file will be saved as %s" % master_file)

    # Get some parameters in here
    if "eiger" in TR.detector_name.lower():
        from .I19_2_params import I19_2Eiger as axes_params

        det_params = EigerDetector(
            "Eiger 2X 4M",
            (2162, 2068),
            "CdTe",
            50649,
            -1,
        )
    elif "tristan" in TR.detector_name.lower():
        from .I19_2_params import I19_2Tristan as axes_params

        det_params = TristanDetector("Tristan 10M", (3043, 4183))
    else:
        raise UnknownDetectorTypeError("Unknown detector name passed.")

    # Get goniometer and detector parameters
    if TR.geometry_json:
        logger.info("Reading geometry from json file.")
        gonio_axes = JSONParamsIO(TR.geometry_json).get_goniometer_axes_from_file()
        det_axes = JSONParamsIO(TR.geometry_json).get_detector_axes_from_file()
        # Overwrite
        axes_params.gonio = gonio_axes
        axes_params.det_axes = det_axes

    if TR.detector_json:
        logger.info("Reading detector parameters from json file.")
        det_params = JSONParamsIO(TR.detector_json).get_detector_params_from_file()
        fast_axis, slow_axis = JSONParamsIO(
            TR.detector_json
        ).get_fast_and_slow_direction_vectors_from_file(det_params.description)
        # Overwrite
        axes_params.fast_axis = fast_axis
        axes_params.slow_axis = slow_axis

    # Get timestamps in the correct format if they aren't already
    _start_time = (
        (
            tr_params["start_time"].strftime("%Y-%m-%dT%H:%M:%S")
            if find_in_dict(tr_params, "start_time")
            else None
        ),
    )  # This should be datetiem type
    _stop_time = (
        (
            tr_params["stop_time"].strftime("%Y-%m-%dT%H:%M:%S")
            if find_in_dict(tr_params, "stop_time")
            else None
        ),
    )  # idem.
    timestamps = (
        get_iso_timestamp(_start_time),
        get_iso_timestamp(_stop_time),
    )

    if "tristan" in TR.detector_name:
        tristan_writer(master_file, TR, axes_params, det_params, timestamps)
    else:
        vds_dtype = (
            np.uint16
            if not find_in_dict(tr_params, "vds_dtype")
            else tr_params["vds_dtype"]
        )
        eiger_writer(master_file, TR, axes_params, det_params, timestamps, vds_dtype)



"""
Create a NeXus file for serial crystallography datasets collected on Eiger detector either on I19-2 or I24 beamlines.
"""

from __future__ import annotations

import logging
from pathlib import Path
from typing import Literal, Tuple, get_args

import numpy as np
from numpy.typing import DTypeLike

from .. import log
from ..nxs_utils import Attenuator, Beam, Detector, EigerDetector, Goniometer, Source
from ..nxs_write.nxmx_writer import NXmxFileWriter
from ..utils import find_in_dict, get_iso_timestamp
from .beamline_utils import (
    BeamlineAxes,
    GeneralParams,
    PumpProbe,
    collection_summary_log,
)

# Define logger
logger = logging.getLogger("nexgen.SSX_Eiger")


EnabledBeamlines = Literal["i24", "i19-2"]
ExperimentTypes = Literal["extruder", "fixed-target"]  # TODO add "3Dgridscan"


class InvalidBeamlineError(Exception):
    def __init__(self, errmsg):
        logger.error(errmsg)


class UnknownExperimentTypeError(Exception):
    def __init__(self, errmsg):
        logger.error(errmsg)


class SerialParams(GeneralParams):
    """Collection parameters for a serial crystallography experiment.

    Args:
        GeneralParams (Basemodel): General collection parameters common to \
            multiple beamlines/experiments, such as exposure time, wavelength, ...
        num_imgs (int): Total number of frames in a collection.
        detector_distance (float): Distance between sample and deterctor, in mm.
        experiment_type (str): Type of collection.
    """

    num_imgs: int
    detector_distance: float
    experiment_type: str


def _define_vds_dtype_from_bit_depth(bit_depth: int) -> DTypeLike:
    """Define dtype of VDS based on the passed bit depth."""
    if bit_depth == 32:
        return np.uint32
    elif bit_depth == 8:
        return np.uint8
    else:
        return np.uint16


def _get_beamline_specific_params(beamline: str) -> Tuple[BeamlineAxes, EigerDetector]:
    """Get beamline specific axes and eiger description.

    Args:
        beamline (str): Beamline name. Allowed values: i24, i19-2.

    Returns:
        Tuple[BeamlineAxes, EigerDetector]: beamline axes description, eiger parameters.
    """
    match beamline.lower():
        case "i24":
            from .I24_params import I24Eiger as axes_params

            eiger_params = EigerDetector(
                "Eiger2 X 9M",
                (3262, 3108),
                "CdTe",
                50649,
                -1,
            )
        case "i19-2":
            from .I19_2_params import I19_2Eiger as axes_params

            eiger_params = EigerDetector(
                "Eiger2 X 4M",
                (2162, 2068),
                "CdTe",
                50649,
                -1,
            )
    return axes_params, eiger_params


def ssx_eiger_writer(
    visitpath: Path | str,
    filename: str,
    beamline: EnabledBeamlines,
    num_imgs: int,
    expt_type: ExperimentTypes = "fixed-target",
    pump_status: bool = False,
    **ssx_params,
):
    """Gather all collection parameters and write the NeXus file for SSX using Eiger detector.

    Args:
        visitpath (Path | str): Collection directory.
        filename (str): Filename root.
        beamline (str): Beamline on which the experiment is being run. Allowed values: i24, i19-2.
        num_imgs (int): Total number of images collected.
        expt_type (str, optional): Experiment type, accepted values: extruder,
            fixed-target, (coming soon: 3Dgridscan). Defaults to "fixed-target".
        pump_status (bool, optional): True for pump-probe experiment. Defaults to False.

    Keyword Args:
        bit_depth (int): bit_depth_image value, from which the vds_dtype is determined. \
            Will default to 32 if not passed.
        exp_time (float): Exposure time, in s.
        det_dist (float): Distance between sample and detector, in mm.
        beam_center (List[float, float]): Beam center position, in pixels.
        transmission (float): Attenuator transmission, in %.
        wavelength (float): Wavelength of incident beam, in A.
        flux (float): Total flux.
        start_time (datetime): Experiment start time.
        stop_time (datetime): Experiment end time.
        chip_info (Dict): For a grid scan, dictionary containing basic chip information.
            At least it should contain: x/y_start, x/y number of blocks and block size, \
            x/y number of steps and number of exposures.
        chipmap (Path | str): Path to the chipmap file corresponding to the experiment,
            if None for a fixed target experiment, it indicates that the fullchip is \
            being scanned.
        pump_exp (float): Pump exposure time, in s.
        pump_delay (float): Pump delay time, in s.
        osc_axis (str): Oscillation axis. Always omega on I24. If not passed it will \
            default to phi for I19-2.
        outdir (str): Directory where to save the file. Only specify if different \
            from meta_file directory.

    Raises:
        InvalidBeamlineError: If an invalid beamline name is passed.
        UnknownExperimentTypeError: If an invalid experiment type is passed.
    """
    # Beamline check
    if beamline.lower() not in get_args(EnabledBeamlines):
        raise InvalidBeamlineError(
            "Unknown beamline for SSX collections with Eiger detector."
            "Beamlines currently enabled for the writer: I24 (Eiger 9M), I19-2 (Eiger 4M)."
        )
    # Collect some of the params
    SSX = SerialParams(
        num_imgs=int(num_imgs),
        exposure_time=(
            ssx_params["exp_time"] if find_in_dict("exp_time", ssx_params) else 0.0
        ),
        detector_distance=(
            ssx_params["det_dist"] if find_in_dict("det_dist", ssx_params) else 0.0
        ),
        experiment_type=expt_type,
        beam_center=(
            ssx_params["beam_center"]
            if find_in_dict("beam_center", ssx_params)
            else (0, 0)
        ),
        wavelength=(
            ssx_params["wavelength"] if find_in_dict("wavelength", ssx_params) else None
        ),
        transmission=(
            ssx_params["transmission"]
            if find_in_dict("transmission", ssx_params)
            else None
        ),
        flux=ssx_params["flux"] if find_in_dict("flux", ssx_params) else None,
    )

    chip_info = (
        ssx_params["chip_info"] if find_in_dict("chip_info", ssx_params) else None
    )
    chipmap = ssx_params["chipmap"] if find_in_dict("chipmap", ssx_params) else None

    if SSX.experiment_type.lower() not in get_args(ExperimentTypes):
        raise UnknownExperimentTypeError(
            f"Unknown experiment type, please pass one of {get_args(ExperimentTypes)}"
        )

    visitpath = Path(visitpath).expanduser().resolve()

    if find_in_dict("outdir", ssx_params) and ssx_params["outdir"]:
        wdir = Path(ssx_params["outdir"]).expanduser().resolve()
    else:
        wdir = visitpath

    # Configure logging
    logfile = wdir / f"{beamline}_EigerSSX_nxs_writer.log"
    log.config(logfile.as_posix())

    logger.info(f"Current collection directory: {visitpath.as_posix()}")
    if wdir != visitpath:
        logger.warning(f"Nexus file will be saved in a different directory: {wdir}")
    # Get NeXus filename
    master_file = wdir / f"{filename}.nxs"
    logger.info("NeXus file will be saved as %s" % master_file.as_posix())

    # Get parameters depending on beamline
    logger.info(f"DLS Beamline: {beamline.upper()}.")
    # Define source
    source = Source(beamline.upper())
    # Axes, eiger params
    axes_params, eiger_params = _get_beamline_specific_params(beamline)
    # Oscillation axis defaults to omega unless it's I19-2
    if beamline.lower() == "i19-2":
        osc_axis = ssx_params["osc_axis"] if "osc_axis" in ssx_params.keys() else "phi"
    else:
        osc_axis = "omega"

    # Define what to do based on experiment type
    logger.info(f"Running {SSX.experiment_type} collection.")

    # Get pump information
    pump_probe = PumpProbe()
    if pump_status is True:
        # Exposure and delay could also be found in dictionary for grid scan
        logger.info("Pump status is True.")
        pump_probe.pump_status = pump_status
        pump_probe.pump_exposure = (
            ssx_params["pump_exp"] if find_in_dict("pump_exp", ssx_params) else None
        )
        pump_probe.pump_delay = (
            ssx_params["pump_exp"] if find_in_dict("pump_exp", ssx_params) else None
        )

        logger.info(f"Recorded pump exposure time: {pump_probe.pump_exposure}")
        logger.info(f"Recorded pump delay time: {pump_probe.pump_delay}")
        if SSX.experiment_type == "fixed-target":
            pump_probe.pump_repeat = int(chip_info["PUMP_REPEAT"][1])

    # Get timestamps in the correct format
    _start_time = (
        ssx_params["start_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if find_in_dict("start_time", ssx_params) and ssx_params["start_time"]
        else None
    )
    _stop_time = (
        ssx_params["stop_time"].strftime("%Y-%m-%dT%H:%M:%S")
        if find_in_dict("stop_time", ssx_params) and ssx_params["stop_time"]
        else None
    )
    timestamps = (
        get_iso_timestamp(_start_time),
        get_iso_timestamp(_stop_time),
    )

    # Define meta file name and check if it has already appeared in the directory
    metafile = visitpath / f"{filename}_meta.h5"
    print(metafile)
    _check = [f for f in visitpath.iterdir() if f.name == metafile.name]
    if len(_check) == 0:
        logger.warning(
            """Meta file has not yet appeared in the visit directory.
            If still missing at the end of the collection, something may be wrong.
            Without a meta file, the links in the nexus file will be broken.
            """
        )
    else:
        logger.debug(f"Found {metafile} in directory.")

    # Define Attenuator
    attenuator = Attenuator(SSX.transmission)
    # Define Beam
    wl = SSX.wavelength
    if not wl:
        logger.warning("No value passed for wavelength, will be set to 0.0.")
        wl = 0.0
    beam = Beam(wl, SSX.flux)

    # Define vds_dtype from bit_depth
    if not find_in_dict("bit_depth", ssx_params):
        logger.warning("Bit depth not in parameters, will be assumed to be 32.")
        bit_depth = 32
    else:
        bit_depth = ssx_params["bit_depth"]
    vds_dtype = _define_vds_dtype_from_bit_depth(bit_depth)
    logger.debug(f"VDS dtype will be {vds_dtype}")

    # Define Goniometer axes
    gonio_axes = axes_params.gonio
    # Define Detector
    det_axes = axes_params.det_axes
    # Set det_z to detector_distance passed in mm
    det_z_idx = [n for n, ax in enumerate(det_axes) if ax.name == "det_z"][0]
    det_axes[det_z_idx].start_pos = SSX.detector_distance

    # Define detector
    detector = Detector(
        eiger_params,
        det_axes,
        SSX.beam_center,
        SSX.exposure_time,
        [axes_params.fast_axis, axes_params.slow_axis],
    )

    tot_num_imgs = SSX.num_imgs

    # Run experiment type
    match SSX.experiment_type:
        case "extruder":
            from .SSX_expt import run_extruder

            gonio_axes, SCAN, pump_info = run_extruder(
                gonio_axes,
                tot_num_imgs,
                pump_probe,
                osc_axis,
            )
        case "fixed-target":
            from .SSX_expt import run_fixed_target

            # Define chipmap if needed
            chipmapfile = (
                None if chipmap is None else Path(chipmap).expanduser().resolve()
            )

            SCAN, pump_info = run_fixed_target(
                gonio_axes,
                chip_info,
                chipmapfile,
                pump_probe,
                ["sam_y", "sam_x"],
            )

            # Sanity check that things make sense
            if SSX.num_imgs != len(SCAN["sam_x"]):
                logger.warning(
                    f"The total number of scan points is {len(SCAN['sam_x'])}, which does not match the total number of images passed as input {SSX.num_imgs}."
                )
                logger.warning(
                    "Reset SSX.num_imgs to number of scan points for vds creation."
                )
                tot_num_imgs = len(SCAN["sam_x"])
        # TODO case "3D"

    # Define goniometer only once the full scan has been calculated.
    goniometer = Goniometer(gonio_axes, SCAN)

    # Log a bunch of stuff
    collection_summary_log(
        logger,
        goniometer,
        detector,
        attenuator,
        beam,
        source,
        timestamps,
    )

    # Get to the actual writing
    try:
        NXmx_Writer = NXmxFileWriter(
            master_file,
            goniometer,
            detector,
            source,
            beam,
            attenuator,
            tot_num_imgs,
        )
        image_filename = metafile.as_posix().replace("_meta.h5", "")
        NXmx_Writer.write(image_filename=image_filename, start_time=timestamps[0])
        if pump_status is True:
            logger.info("Write pump information to file.")
            NXmx_Writer.add_NXnote(
                notes=pump_info,
                loc="/entry/source/notes",
            )
        NXmx_Writer.update_timestamps(timestamps[1], "end_time")
        NXmx_Writer.write_vds(
            vds_shape=(tot_num_imgs, *detector.detector_params.image_size),
            vds_dtype=vds_dtype,
        )
        logger.info(f"The file {master_file} was written correctly.")
    except Exception as err:
        logger.exception(err)
        logger.info(
            f"An error occurred and {master_file} couldn't be written correctly."
        )
        raise



"""
Define and store basic beamline utilities.
"""

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple

from pydantic import BaseModel

from nexgen.nxs_utils import Attenuator, Axis, Beam, Detector, Goniometer, Source
from nexgen.utils import Point3D


class GeneralParams(BaseModel):
    """Parameters passed as input from the beamline.

    Args:
        exposure_time (float): Exposure time, in s.
        beam_center (Sequence[float]): Beam center (x,y) position, in pixels.
        wavelength (float, optional): Incident beam wavelength, in A.
        transmission (float, optional): Attenuator transmission, in %.
        flux (float, optional): Total flux.
    """

    exposure_time: float
    beam_center: Sequence[float]
    wavelength: Optional[float]
    transmission: Optional[float]
    flux: Optional[float]


class PumpProbe(BaseModel):
    """
    Define pump probe parameters for a serial experiment on I24.

    Args:
        pump_status (bool, optional): Pump probe on/off.
        pump_exposure (float, optional): Pump exposure time, in s.
        pump_delay (float, optional): Pump delay, in s.
        pump_repeat (int, optional): Repeat mode.
    """

    pump_status: Optional[bool] = False
    pump_exposure: Optional[float] = None
    pump_delay: Optional[float] = None
    pump_repeat: Optional[int] = 0


@dataclass
class BeamlineAxes:
    """Beamline specific axes for goniometer, detector and detector module."""

    gonio: List[Axis]
    det_axes: List[Axis]
    fast_axis: Point3D | Tuple[float, float, float]
    slow_axis: Point3D | Tuple[float, float, float]

    def __post_init__(self):
        if not isinstance(self.fast_axis, Point3D):
            self.fast_axis = Point3D(*self.fast_axis)
        if not isinstance(self.slow_axis, Point3D):
            self.slow_axis = Point3D(*self.slow_axis)


def collection_summary_log(
    logger: logging.Logger,
    goniometer: Goniometer,
    detector: Detector,
    attenuator: Attenuator,
    beam: Beam,
    source: Source,
    timestamps: Tuple[str],
):
    """General function to log a collection summary."""
    logger.debug("--- COLLECTION SUMMARY ---")
    logger.debug(source.__repr__())

    logger.debug(f"Incident beam wavelength: {beam.wavelength}")
    logger.debug(f"Attenuation: {attenuator.transmission}")

    logger.debug(goniometer.__repr__())
    logger.debug(detector.__repr__())

    logger.debug(f"Recorded beam center is: {detector.beam_center}.")
    logger.debug(f"Recorded exposure time: {detector.exp_time} s.")

    logger.debug(f"Timestamps recorded: {timestamps}")



"""
A writer for NXmx-like NeXus files for electron diffraction.
"""

from __future__ import annotations

import logging
from datetime import datetime
from pathlib import Path
from typing import Dict, List

import h5py
import numpy as np
from numpy.typing import DTypeLike

from ..nxs_utils import Attenuator, Beam, Detector, Goniometer, Source
from ..tools.vds_tools import image_vds_writer, vds_file_writer
from ..utils import coord2mcstas
from .nxclass_writers import (
    write_NXcoordinate_system_set,
    write_NXdata,
    write_NXdatetime,
    write_NXdetector,
    write_NXdetector_module,
    write_NXentry,
    write_NXinstrument,
    write_NXsample,
    write_NXsource,
)
from .write_utils import calculate_estimated_end_time

# Logger
edwriter_logger = logging.getLogger("nexgen.EDNXmxFileWriter")
edwriter_logger.setLevel(logging.DEBUG)


class EDNXmxFileWriter:
    """A class to generate NXmx-like NeXus files for electron diffraction.

    Requires a couple of additional arguments compared to a standard NXmxWriter:
        ED_coord_system (Dict): Definition of the current coordinate frame for ED. \
            It should at least contain the convention, origin and base vectors.
        convert_to_mcstas (bool, optional): If true, convert the vectors to mcstas. \
            Defaults to False.
    """

    def __init__(
        self,
        filename: Path | str,
        goniometer: Goniometer,
        detector: Detector,
        source: Source,
        beam: Beam,
        attenuator: Attenuator,
        tot_num_imgs: int,
        ED_coord_system: Dict,
        convert_to_mcstas: bool = False,
    ):
        self.filename = Path(filename).expanduser().resolve()
        self.goniometer = goniometer
        self.detector = detector
        self.source = source
        self.beam = beam
        self.attenuator = attenuator
        self.tot_num_imgs = tot_num_imgs
        self.ED_coord_system = ED_coord_system
        self.convert_cs = convert_to_mcstas

    def _check_coordinate_frame(self):
        """Checks the coordinate frame and converts to mcstas if requested."""
        if self.convert_cs is True:
            edwriter_logger.warning(
                "All the vector/offset axis coordinates will be converted to mcstas."
            )
            mat = np.array(
                [
                    self.ED_coord_system["x"].vector,
                    self.ED_coord_system["y"].vector,
                    self.ED_coord_system["z"].vector,
                ]
            )
            # TODO: Need to redefine ED as {"x": Axis()} or something
            for ax in self.goniometer.axes_list:
                ax.vector = coord2mcstas(ax.vector, mat)
            for ax in self.detector.detector_axes:
                ax.vector = coord2mcstas(ax.vector, mat)
            self.detector.fast_axis = coord2mcstas(self.detector.fast_axis, mat)
            self.detector.slow_axis = coord2mcstas(self.detector.slow_axis, mat)

    def _get_data_filename(self) -> List[Path]:
        """Returns a list of data files."""
        image_filename = f"{self.filename.stem}_data_000001.h5"
        return self.filename.parent / f"{image_filename}"

    def _get_collection_time(self) -> float:
        """Returns total collection time."""
        return self.detector.exp_time * self.tot_num_imgs

    def write(
        self,
        image_datafiles: List | None = None,
        data_entry_key: str = "/entry/data/data",
        start_time: datetime | str | None = None,
        write_mode: str = "x",
    ):
        """Write a NXmx-like NeXus file for electron diffraction.

        This function calls the writers for the main NXclass objects.
        Additionally, it performs a few checks on the coordinate frame of the input vectors \
        and then calls the writers for the relevant NeXus base classes.

        Args:
            image_datafiles (List | None, optional): List of image data files. If not \
                passed, the program will look for files with the stem_data_######.h5 in \
                the target directory. Defaults to None.
            data_entry_key (str, optional): Dataset entry key in datafiles. Defaults to \
                entry/data/data.
            start_time (datetime | str, optional): Collection estimated end time if \
                available, in the format "%Y-%m-%dT%H:%M:%SZ". Defaults to None.
            write_mode (str, optional): String indicating writing mode for the output \
                NeXus file. Accepts any valid h5py file opening mode. Defaults to "x".
        """
        # Get data files
        datafiles = image_datafiles if image_datafiles else [self._get_data_filename()]

        module = self.detector.get_module_info()

        # Scans
        osc, _ = self.goniometer.define_scan_from_goniometer_axes()

        # Deal with vecotrs/offsets/whatever if needed
        self._check_coordinate_frame()

        # Get the instrument name
        instrument_name = self.source.set_instrument_name()
        edwriter_logger.info(f"Instrument name will be set as {instrument_name}.")

        # NXcoordinate_system_set: /entry/coordinate_system_set
        base_vectors = {k: self.ED_coord_system.get(k) for k in ["x", "y", "z"]}

        with h5py.File(self.filename, write_mode) as nxs:
            # NXentry and NXmx definition
            write_NXentry(nxs)

            write_NXcoordinate_system_set(
                nxs,
                self.ED_coord_system["convention"],
                base_vectors,
                self.ED_coord_system["origin"],
            )

            if start_time:
                write_NXdatetime(nxs, start_time, "start_time")
                tot_exp_time = self._get_collection_time()
                est_end = calculate_estimated_end_time(start_time, tot_exp_time)
                write_NXdatetime(nxs, est_end, "end_time_estimated")

            # NXdata: entry/data
            write_NXdata(
                nxs,
                datafiles,
                "images",
                list(osc.keys())[0],
                entry_key=data_entry_key,
            )

            # NXinstrument: entry/instrument
            write_NXinstrument(
                nxs,
                self.beam,
                self.attenuator,
                self.source,
                reset_instrument_name=True,
            )

            # NXdetector: entry/instrument/detector
            write_NXdetector(
                nxs,
                self.detector,
                self.tot_num_imgs,
            )

            # NXmodule: entry/instrument/detector/module
            write_NXdetector_module(
                nxs,
                module,
                self.detector.detector_params.image_size,
                self.detector.detector_params.pixel_size,
                beam_center=self.detector.beam_center,
            )

            # NXsource: entry/source
            write_NXsource(nxs, self.source)

            # NXsample: entry/sample
            write_NXsample(
                nxs,
                self.goniometer.axes_list,
                "images",
                osc,
            )

    def write_vds(
        self,
        vds_dtype: DTypeLike = np.uint16,
        writer_type: str = "dataset",
        data_entry_key: str = "/entry/data/data",
        datafiles: List[Path] | None = None,
    ):
        """Write a vds for electron diffraction.

        This method adds a VDS under /entry/data/data in the NeXus file as a default.
        If instead the writer_type is set to "file" it will write an external _vds.h5 file.

        Args:
            vds_dtype (DTypeLike, optional): The type of the input data. Defaults to np.uint16.
            writer_type (str, optional): Type of vds required. Defaults to "dataset".
            data_entry_key (str, optional): Dataset entry key in datafiles. Defaults to "/entry/data/data".
            datafiles ((List | None, optional): List of image data files. If not passed, the program will look for \
                files with the stem_######.h5 in the target directory. Defaults to None.
        """
        if writer_type not in ["dataset", "file"]:
            edwriter_logger.warning(
                f"Writer type {writer_type} unknown. Will default to dataset."
            )
            writer_type = "dataset"
        with h5py.File(self.filename, "r+") as nxs:
            if writer_type == "dataset":
                edwriter_logger.debug(
                    "Writing vds dataset as /entry/data/data in nexus file."
                )
                image_vds_writer(
                    nxs,
                    (self.tot_num_imgs, *self.detector.detector_params.image_size),
                    data_type=vds_dtype,
                    entry_key=data_entry_key,
                )
            else:
                if not datafiles:
                    edwriter_logger.warning(
                        "No datafile list passed, vds file won't be written."
                    )
                    return
                edwriter_logger.info(
                    "Writing external vds file with link in /entry/data/data in nexus file."
                )
                vds_file_writer(
                    nxs,
                    datafiles,
                    (self.tot_num_imgs, *self.detector.detector_params.image_size),
                    vds_dtype,
                    data_entry_key,
                )



"""
A script to convert a series of individual MRC images into a Nexus format.
"""

import logging
import os
from argparse import ArgumentParser
from pathlib import Path

import numpy as np

from nexgen.beamlines.ED_params import ED_coord_system, EDCeta, EDSource
from nexgen.nxs_utils import Attenuator, Beam, CetaDetector, Detector, Goniometer
from nexgen.nxs_utils.scan_utils import calculate_scan_points
from nexgen.nxs_write.ed_nxmx_writer import EDNXmxFileWriter
from nexgen.tools.mrc_tools import collect_data, get_metadata

logger = logging.getLogger("nexgen.ED_mrc_to_nexus")


def main():

    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "--det_distance",
        type=float,
        default=None,
        help="The sample-detector distance.",
    )
    parser.add_argument(
        "-wl",
        "--wavelength",
        type=float,
        default=None,
        help="Incident beam wavelength, in A.",
    )
    parser.add_argument(
        "--axis-start",
        type=float,
        default=None,
        help="Rotation axis start position.",
    )
    parser.add_argument(
        "--axis-inc",
        type=float,
        default=None,
        help="Rotation axis increment.",
    )
    parser.add_argument(
        "-e",
        "--exp-time",
        type=float,
        help="Exposure time, in s.",
    )
    parser.add_argument(
        "-bc",
        "--beam-center",
        type=float,
        nargs=2,
        help="Beam center (x,y) positions.",
    )
    parser.add_argument("input_files", nargs="+", help="List of input files")

    args = parser.parse_args()

    logger.info("Starting MRC to Nexus conversion")

    for file in args.input_files:
        if not file.endswith(".mrc"):
            raise ValueError("Not an MRC file: %s" % file)

    mrc_files = args.input_files
    full_mrc_path = os.path.abspath(mrc_files[0])
    logger.info("Collecting MRC data into a h5 file")
    tot_imgs, out_file, angles = collect_data(mrc_files)

    mdict = get_metadata(mrc_files[0])

    attenuator = Attenuator(transmission=None)

    if args.det_distance is not None:
        det_distance = args.det_distance
    elif "cameraLength" in mdict:
        det_distance = mdict["cameraLength"]
        msg = "Reading detector distance from "
        msg += "the MRC files: %.1f" % det_distance
        logger.info(msg)
    else:
        msg = "No detector distance in the MRC metadata. "
        msg += "You can set it with -det_distance option."
        raise ValueError(msg)

    if args.wavelength is not None:
        beam = Beam(args.wavelength)
    elif "wavelength" in mdict:
        beam = Beam(mdict["wavelength"])
        msg = "Reading wavelength from the MRC files: %f" % mdict["wavelength"]
        logger.info(msg)
    else:
        msg = "No wavelength in the MRC metadata. "
        msg += "You can set it with --wavelength option."
        raise ValueError(msg)

    if args.axis_start is not None:
        start_angle = args.axis_start
    elif angles[0] is not None:
        start_angle = angles[0]
        msg = "Reading starting angle from the MRC files: %.2f" % start_angle
        logger.info(msg)
    else:
        msg = "No starting angle in the MRC metadata. "
        msg += "You can set it with --axis_start option."
        raise ValueError(msg)

    if args.axis_inc is not None:
        increment = args.axis_inc
    elif "tiltPerImage" in mdict:
        increment = mdict["tiltPerImage"]
        msg = "Reading angle increment from the MRC files: %f" % increment
        logger.info(msg)
    else:
        msg = "No angle increment in the MRC metadata. "
        msg += "You can set it with --axis_inc option."
        raise ValueError(msg)

    if args.exp_time is not None:
        exposure_time = args.exp_time
    elif "integrationTime" in mdict:
        exposure_time = mdict["integrationTime"]
        msg = "Reading exposure time from the MRC files: %f" % exposure_time
        logger.info(msg)
    else:
        msg = "No exposure time in the MRC metadata. "
        msg += "You can set it with --exp_time option."
        raise ValueError(msg)

    if args.beam_center is not None:
        if len(args.beam_center) == 2:
            x, y = args.beam_center
            beam_center = (x, y)
        else:
            msg = "Beam center requires two arguments"
            raise ValueError(msg)
    elif ("beamCentreXpx" in mdict) and ("beamCentreYpx" in mdict):
        beam_center = (mdict["beamCentreXpx"], mdict["beamCentreYpx"])
        msg = "Reading beam center from the MRC files: (%.1f, %.1f)" % beam_center
        logger.info(msg)
    else:
        msg = "No beam center in the MRC metadata. "
        msg += "You can set it with --beam_center option."
        raise ValueError(msg)

    gonio_axes = EDCeta.gonio
    gonio_axes[0].start_pos = start_angle
    gonio_axes[0].increment = increment
    gonio_axes[0].num_steps = tot_imgs

    scan = calculate_scan_points(gonio_axes[0], rotation=True, tot_num_imgs=tot_imgs)
    goniometer = Goniometer(gonio_axes, scan)

    nx = mdict["nx"]
    ny = mdict["ny"]

    if (nx == 2048) and (ny == 2048):
        binning = 2
    else:
        binning = 1

    det_params = CetaDetector("Thermo Fisher Ceta-D", [nx, ny])

    pixel_size = 0.014 * binning
    ps_str = "%5.3fmm" % pixel_size
    det_params.pixel_size = [ps_str, ps_str]
    det_params.overload = 8000 * binning**2

    extra_params = {}
    mask = np.zeros((nx, ny), dtype=np.int32)
    flatfield = np.ones((nx, ny), dtype=np.float32)
    extra_params["pixel_mask"] = mask
    extra_params["flatfield"] = flatfield
    extra_params["pixel_mask_applied"] = 1
    extra_params["flatfield_applied"] = 1
    det_params.constants.update(extra_params)

    det_axes = EDCeta.det_axes

    det_axes[0].start_pos = det_distance

    detector = Detector(
        det_params,
        det_axes,
        beam_center,
        exposure_time,
        [EDCeta.fast_axis, EDCeta.slow_axis],
    )

    source = EDSource
    source.beamline = "Ceta"

    script_dir = os.path.dirname(full_mrc_path)
    file_name = out_file.replace("h5", "nxs")
    full_path = os.path.join(script_dir, file_name)
    data_path = os.path.join(script_dir, out_file)
    data_path = Path(data_path)

    msg = "Writing the Nexus file %s" % full_path
    logger.info(msg)
    writer = EDNXmxFileWriter(
        full_path,
        goniometer,
        detector,
        source,
        beam,
        attenuator,
        tot_imgs,
        ED_coord_system,
    )

    datafiles = [data_path]
    writer.write(datafiles, "/entry/data/data")
    writer.write_vds(vds_dtype=np.int32, datafiles=datafiles)
    logger.info("MRC images converted to Nexus.")



""" Helper functions for the MRC to Nexus converter """

import os
from math import sqrt

import h5py
import hdf5plugin
import mrcfile
import numpy as np


def cal_wavelength(V0):
    """
    Compute the relativistic electron wavelength from
    the electron acceleration voltage.

    Args:
        V0 (int or float): acceleration voltage for electrons
                           in Volts [V].

    Returns:
        wlen (float): Electron wavelength in Angstroms.

    For an acceleration voltage of 200000 V, the electron energy
    is 200 keV (set by default if V0 is zero).
    """

    h = 6.626e-34  # Planck's constant [J*s]
    m = 9.109e-31  # Electron mass [kg]
    e = 1.6021766208e-19  # Electron charge [C]
    c = 3e8  # Speed of light  [m/s]

    # Default to wavelength at 200 keV if voltage set to zero
    if V0 == 0:
        V0 = 200000
    wlen = h / sqrt(2 * m * e * V0 * (1 + e * V0 / (2 * m * c * c))) * 1e10
    return wlen  # Electron wavelength in Angstroms


def get_metadata(mrc_image, verbatim=True):
    """
    Extract the metadata dictionary from a single MRC file.

    Args:
        mrc_image (Path or str): the MRC file path.

    Returns:
        hd (dict): A dictionary containing metadata
                   from the MRC file.

    The function extracts data from the MRC header and extended
    header dictionaries.
    """

    with mrcfile.open(mrc_image, header_only=True) as mrc:
        h = mrc.header
        try:
            xh = mrc.indexed_extended_header
        except AttributeError:
            # For mrcfile versions older than 1.5.0
            xh = mrc.extended_header
        hd = {}

        hd["nx"] = h["nx"]
        hd["ny"] = h["ny"]
        hd["nz"] = h["nz"]
        hd["mx"] = h["mx"]
        hd["my"] = h["my"]
        hd["mz"] = h["mz"]

        h_ext_type = h["exttyp"].item()
        is_FEI = h_ext_type.startswith(b"FEI")
        is_FEI2 = h_ext_type.startswith(b"FEI2")

        if not is_FEI:
            return hd

        hd["alphaTilt"] = xh["Alpha tilt"][0]
        hd["integrationTime"] = xh["Integration time"][0]
        hd["tilt_axis"] = xh["Tilt axis angle"][0]
        hd["pixelSpacing"] = xh["Pixel size X"][0]

        if hd["pixelSpacing"] == 0:
            raise ValueError("Incorrect extended header")

        hd["acceleratingVoltage"] = xh["HT"][0]
        hd["camera"] = xh["Camera name"][0]
        hd["binning"] = xh["Binning Width"][0]
        hd["noiseReduction"] = xh["Ceta noise reduction"][0]
        hd["physicalPixel"] = 14e-6
        hd["wavelength"] = cal_wavelength(hd["acceleratingVoltage"])
        hd["cameraLength"] = (
            (hd["physicalPixel"] * hd["binning"])
            / (hd["pixelSpacing"] * hd["wavelength"] * 1e-10)
            * 1000.0
        )

        if not is_FEI2:
            return hd

        hd["scanRotation"] = xh["Scan rotation"][0]
        hd["diffractionPatternRotation"] = xh["Diffraction pattern rotation"][0]
        hd["imageRotation"] = xh["Image rotation"][0]
        hd["scanModeEnum"] = xh["Scan mode enumeration"][0]
        hd["acquisitionTimeStamp"] = xh["Acquisition time stamp"][0]
        hd["detectorCommercialName"] = xh["Detector commercial name"][0]
        hd["startTiltAngle"] = xh["Start tilt angle"][0]
        hd["endTiltAngle"] = xh["End tilt angle"][0]
        hd["tiltPerImage"] = xh["Tilt per image"][0]
        hd["tiltSpeed"] = xh["Tilt speed"][0]
        hd["beamCentreXpx"] = xh["Beam center X pixel"][0]
        hd["beamCentreYpx"] = xh["Beam center Y pixel"][0]
        hd["cfegFlashTimestamp"] = xh["CFEG flash timestamp"][0]
        hd["phasePlatePositionIndex"] = xh["Phase plate position index"][0]
        hd["objectiveApertureName"] = xh["Objective aperture name"][0]

        # Check if binning is correct
        assert hd["binning"] == 4096 / hd["nx"]

        return hd


def collect_data(files):
    """
    Extract image data from a list of MRC files into a single
    HDF5 (h5) file.

    Args:
        files (list of strings): A list containing names
            of all MRC images.

    Returns:
        n (int): The total number of MRC images.
        out_file (string): Name of the output h5 file.
        angles (1D numpy array of floats): An array containing
            tilt angles of all MRC images (as extracted from the
            MRC metadata).

    The function takes a list of MRC filenames, assuming that each
    filename contains a single diffraction image (as a 2D array),
    and makes a single h5 file containing a 3D array (where the first
    index enumerates the images). The data is saved in the field
    '/entry/data/data/' as an int32 and compressed using the LZ4
    compression method. Note that the original data in the MRC files
    is saved as a floating point number, although the actual data
    consists solely of integers. The function also assumes that all
    MRC files end with an indexing number: like 'basename_00001.mrc',
    in which case the data is saved in 'basename.h5'.
    """

    out_file = files[0].rsplit("_", 1)[0] + ".h5"
    mrc_files = []
    angles = []
    for file in files:
        path = os.path.abspath(file)
        if path.endswith(".mrc"):
            mrc_files.append(file)

    n = len(mrc_files)

    test_file = mrcfile.open(mrc_files[0], mode="r")
    data_shape = test_file.data.shape
    test_file.close()
    if len(data_shape) != 2:
        msg = "The converter works only with individual MRC images"
        raise ValueError(msg)

    with h5py.File(out_file, "w") as hdf5_file:
        dataset_shape = (n, data_shape[0], data_shape[1])
        dataset = hdf5_file.create_dataset(
            "data_temp", shape=dataset_shape, dtype=np.int32
        )

        group = hdf5_file.create_group("entry")
        group.attrs["NX_class"] = np.bytes_("NXentry")
        data_group = group.create_group("data")
        data_group.attrs["NX_class"] = np.bytes_("NXdata")

        compressed_data = hdf5_file.create_dataset(
            "/entry/data/data", shape=dataset_shape, dtype=np.int32, **hdf5plugin.LZ4()
        )

        for i, file in enumerate(mrc_files):
            print(" Reading image %d:  %s" % (i, file))
            mrc = mrcfile.open(file, mode="r")
            data = np.array(mrc.data, dtype=np.int32)
            try:
                xh = mrc.indexed_extended_header
            except AttributeError:
                xh = mrc.extended_header

            if "Alpha tilt" in xh.dtype.names:
                angles.append(xh["Alpha tilt"][0])
            else:
                angles.append(None)
            dataset[i, :, :] = data
            mrc.close()

        compressed_data[...] = dataset[...]
        del hdf5_file["data_temp"]

        return n, out_file, np.array(angles)



"""
A writer for NXmx format NeXus Files.
"""

from __future__ import annotations

import logging
import math
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Tuple

import h5py
import numpy as np
from numpy.typing import DTypeLike

from ..nxs_utils.detector import Detector
from ..nxs_utils.goniometer import Goniometer
from ..nxs_utils.source import Attenuator, Beam, Source
from ..tools.vds_tools import (
    clean_unused_links,
    image_vds_writer,
    jungfrau_vds_writer,
)
from ..utils import (
    MAX_FRAMES_PER_DATASET,
    MAX_SUFFIX_DIGITS,
    get_filename_template,
)
from .nxclass_writers import (
    write_NXdata,
    write_NXdatetime,
    write_NXdetector,
    write_NXdetector_module,
    write_NXentry,
    write_NXinstrument,
    write_NXnote,
    write_NXsample,
    write_NXsource,
)
from .write_utils import TSdset, calculate_estimated_end_time

# Logger
nxmx_logger = logging.getLogger("nexgen.NXmxFileWriter")
nxmx_logger.setLevel(logging.DEBUG)


class NXmxFileWriter:
    """A class to generate NXmx format NeXus files."""

    def __init__(
        self,
        filename: Path | str,
        goniometer: Goniometer,
        detector: Detector,
        source: Source,
        beam: Beam,
        attenuator: Attenuator,
        tot_num_imgs: int,  # | None = None,
    ):
        self.filename = Path(filename).expanduser().resolve()
        self.goniometer = goniometer
        self.detector = detector
        self.source = source
        self.beam = beam
        self.attenuator = attenuator
        self.tot_num_imgs = tot_num_imgs

    def _get_meta_file(self, image_filename: str = None) -> Path | None:
        """Get filename_meta.h5 file in directory if it's supposed to exist."""
        if self.detector.detector_params.hasMeta is False:
            nxmx_logger.debug("No meta file for this collection.")
            return None
        if image_filename:
            return self.filename.parent / f"{image_filename}_meta.h5"
        else:
            return self.filename.parent / f"{self.filename.stem}_meta.h5"

    def _get_collection_time(self) -> float:
        """_Returns total collection time."""
        return self.detector.exp_time * self.tot_num_imgs

    def _get_data_files_list(
        self,
        image_filename: str | None = None,
    ) -> List[Path]:
        """Get list of datafiles.

        Args:
            image_filename (str | None, optional): Filename stem to use to look for image files. Needed in case it doesn't match \
                the NeXus file name. Defaults to None.

        Returns:
            List[Path]: List of data files to link to.
        """
        num_files = math.ceil(self.tot_num_imgs / MAX_FRAMES_PER_DATASET)
        template = (
            get_filename_template(self.filename)
            if not image_filename
            else (
                self.filename.parent / f"{image_filename}_%0{MAX_SUFFIX_DIGITS}d.h5"
            ).as_posix()
        )
        datafiles = [Path(template % i) for i in range(1, num_files + 1)]
        nxmx_logger.debug(f"Number of datafiles to be written: {len(datafiles)}.")
        return datafiles

    def update_timestamps(
        self, timestamp: datetime | str, dset_name: TSdset = "end_time"
    ):
        """Save timestamps for start and/or end collection if not written before.

        Args:
            timestamp (datetime | str): Timestamp, as datetime or str.
            dset_name (TSdset, optional): Name of dataset to write to nexus file.\
                Allowed values: ["start_time", "end_time", "end_time_estimated". Defaults to "end_time".
        """
        with h5py.File(self.filename, "r+") as nxs:
            write_NXdatetime(nxs, timestamp, dset_name)
        nxmx_logger.info(f"{dset_name} timestamp for collection updated.")

    def add_NXnote(self, notes: Dict, loc: str = "/entry/notes"):
        """Save any additional information as NXnote at the end of the collection.

        Args:
            notes (Dict): Dictionary of (key, value) pairs where key represents the \
                dataset name and value its data.
            loc (str, optional): Location in the NeXus file to save metadata. \
                Defaults to "/entry/notes".
        """
        with h5py.File(self.filename, "r+") as nxs:
            write_NXnote(nxs, loc, notes)
        nxmx_logger.debug(f"Notes saved in {loc}.")

    def write(
        self,
        image_datafiles: List | None = None,
        image_filename: str | None = None,
        start_time: datetime | str | None = None,
        est_end_time: datetime | str | None = None,
        write_mode: str = "x",
        add_non_standard: bool = True,
    ):
        """Write the NXmx format NeXus file.

        This function calls the writers for the main NXclass objects.

        Args:
            image_datafiles (List | None, optional): List of image data files. If not passed, \
                the program will look for files with the stem_######.h5 in the target directory. \
                Defaults to None.
            image_filename (str | None, optional): Filename stem to use to look for image files. \
                Needed in case it doesn't match the NeXus file name. Format: filename_runnumber. \
                Defaults to None.
            start_time (datetime | str, optional): Collection start time if available, in the \
                format "%Y-%m-%dT%H:%M:%SZ".Defaults to None.
            est_end_time (datetime | str, optional): Collection estimated end time if available, \
                in the format "%Y-%m-%dT%H:%M:%SZ". Defaults to None.
            write_mode (str, optional): String indicating writing mode for the output NeXus file. \
                Accepts any valid h5py file opening mode. Defaults to "x".
            add_non_standard (bool, optional): Flag if non-standard NXsample fields should be added \
                for processing to work. Defaults to True, will change in the future.
        """
        metafile = self._get_meta_file(image_filename)
        if metafile:
            nxmx_logger.debug(f"Metafile name: {metafile.as_posix()}.")

        datafiles = (
            image_datafiles
            if image_datafiles
            else self._get_data_files_list(image_filename)
        )

        module = self.detector.get_module_info()

        osc, transl = self.goniometer.define_scan_from_goniometer_axes()

        with h5py.File(self.filename, write_mode) as nxs:
            # NXentry and NXmx definition
            write_NXentry(nxs)

            # Start and estimated end time if known
            if start_time:
                write_NXdatetime(nxs, start_time, "start_time")
                if not est_end_time:
                    tot_exp_time = self._get_collection_time()
                    est_end = calculate_estimated_end_time(start_time, tot_exp_time)
                    write_NXdatetime(nxs, est_end, "end_time_estimated")
            if est_end_time:
                write_NXdatetime(nxs, est_end_time, "end_time_estimated")

            # NXdata: entry/data
            write_NXdata(
                nxs,
                datafiles,
                "images",
                list(osc.keys())[0],
            )

            # NXinstrument: entry/instrument
            write_NXinstrument(
                nxs,
                self.beam,
                self.attenuator,
                self.source,
            )

            # NXdetector: entry/instrument/detector
            write_NXdetector(
                nxs,
                self.detector,
                self.tot_num_imgs,
                metafile,
            )

            # NXmodule: entry/instrument/detector/module
            write_NXdetector_module(
                nxs,
                module,
                self.detector.detector_params.image_size,
                self.detector.detector_params.pixel_size,
                beam_center=self.detector.beam_center,
            )

            # NXsource: entry/source
            write_NXsource(nxs, self.source)

            # NXsample: entry/sample
            write_NXsample(
                nxs,
                self.goniometer.axes_list,
                "images",
                osc,
                transl,
                sample_depends_on=None,  # TODO
                add_nonstandard_fields=add_non_standard,
            )

    def write_vds(
        self,
        vds_offset: int = 0,
        vds_shape: Tuple[int, int, int] = None,
        vds_dtype: DTypeLike = np.uint16,
        clean_up: bool = False,
    ):
        """Write a Virtual Dataset.

        This method adds a VDS under /entry/data/data in the NeXus file, linking to either the full datasets or the subset defined by \
        vds_offset (used as start index) and vds_shape.
        WARNING. Only use clean up if the data collection is finished and all the files have already been written.

        Args:
            vds_offset (int, optional): Start index for the vds writer. Defaults to 0.
            vds_shape (Tuple[int,int,int], optional): Shape of the data which will be linked in the VDS. If not passed, it will be defined as \
            (tot_num_imgs - start_idx, *image_size). Defaults to None.
            vds_dtype (DTypeLike, optional): The type of the input data. Defaults to np.uint16.
            clean_up(bool, optional): Clean up unused links in vds. Defaults to False.
        """
        if not vds_shape:
            vds_shape = (
                self.tot_num_imgs - vds_offset,
                *self.detector.detector_params.image_size,
            )
            if self.goniometer.get_number_of_scan_points() != vds_shape[0]:
                vds_shape = (
                    self.goniometer.get_number_of_scan_points(),
                    *vds_shape[1:],
                )
                nxmx_logger.warning(
                    "The number of scan points doesn't match the calculated vds_shape. \
                    Resetting it to match the number of frames indicated by the scan."
                )

        nxmx_logger.debug(f"VDS shape set to {vds_shape}.")

        with h5py.File(self.filename, "r+") as nxs:
            if "jungfrau" in self.detector.detector_params.description.lower():
                jungfrau_vds_writer(
                    nxs,
                    (self.tot_num_imgs, *self.detector.detector_params.image_size),
                    data_type=vds_dtype,
                )
            else:
                image_vds_writer(
                    nxs,
                    (self.tot_num_imgs, *self.detector.detector_params.image_size),
                    start_index=vds_offset,
                    vds_shape=vds_shape,
                    data_type=vds_dtype,
                )
            if clean_up is True:
                nxmx_logger.warning("Starting clean up of unused links.")
                clean_unused_links(
                    nxs,
                    vds_shape=vds_shape,
                    start_index=vds_offset,
                )

            # If number of frames in the VDS is lower than the total, nimages in NXcollection should be overwritten to match this
            if vds_shape[0] < self.tot_num_imgs:
                del nxs["/entry/instrument/detector/detectorSpecific/nimages"]
                nxs["/entry/instrument/detector/detectorSpecific"].create_dataset(
                    "nimages", data=vds_shape[0]
                )


class EventNXmxFileWriter(NXmxFileWriter):
    """A class to generate NXmx-like NeXus files for event mode data."""

    def __init__(
        self,
        filename: Path | str,
        goniometer: Goniometer,
        detector: Detector,
        source: Source,
        beam: Beam,
        attenuator: Attenuator,
        axis_end_position: float | None = None,
    ):
        super().__init__(
            filename,
            goniometer,
            detector,
            source,
            beam,
            attenuator,
            None,
        )
        self.end_pos = axis_end_position

    def write(
        self,
        start_time: datetime | str | None = None,
        write_mode: str = "x",
        add_non_standard: bool = False,
    ):
        """Write a NXmx-like NeXus file for event mode data collections.

        This method overrides the write() method of NXmxFileWriter, from which thsi class inherits.

        Args:
            start_time (datetime | str, optional): Collection estimated end time if available, in the format "%Y-%m-%dT%H:%M:%SZ".\
                Defaults to None.
            write_mode (str, optional): String indicating writing mode for the output NeXus file. Accepts any valid \
                h5py file opening mode. Defaults to "x".
            add_non_standard (bool, optional): Flag if non-standard NXsample fields should be added \
                for processing to work. Defaults to False.
        """
        # Get metafile
        # No data files, just link to meta
        metafile = super()._get_meta_file()

        module = self.detector.get_module_info()

        # Get scan info
        # Here no scan, just get (start, stop) from omega/phi as osc and None as transl
        osc, _ = self.goniometer.define_scan_axes_for_event_mode(self.end_pos)

        with h5py.File(self.filename, write_mode) as nxs:
            # NXentry and NXmx definition
            write_NXentry(nxs)

            # Write start time if known
            if start_time:
                write_NXdatetime(nxs, start_time, "start_time")

            # NXdata: entry/data
            write_NXdata(
                nxs,
                [metafile],
                "events",
                list(osc.keys())[0],
            )

            # NXinstrument: entry/instrument
            write_NXinstrument(
                nxs,
                self.beam,
                self.attenuator,
                self.source,
            )

            # NXdetector: entry/instrument/detector
            write_NXdetector(
                nxs,
                self.detector,
                meta=metafile,
            )

            # NXmodule: entry/instrument/detector/module
            write_NXdetector_module(
                nxs,
                module,
                self.detector.detector_params.image_size,
                self.detector.detector_params.pixel_size,
                beam_center=self.detector.beam_center,
            )

            # NXsource: entry/source
            write_NXsource(nxs, self.source)

            # NXsample: entry/sample
            write_NXsample(
                nxs,
                self.goniometer.axes_list,
                "events",
                osc,
                sample_depends_on=None,  # TODO
                add_nonstandard_fields=add_non_standard,
            )



"""
Utilities for writing new NeXus format files.
"""

from __future__ import annotations

import logging
import math
from datetime import datetime, timedelta
from pathlib import Path
from typing import List, Literal, Tuple

import h5py  # isort: skip
import numpy as np
from hdf5plugin import Bitshuffle, Blosc
from numpy.typing import ArrayLike

from ..nxs_utils import Axis

# Logger
NXclassUtils_logger = logging.getLogger("nexgen.NXclass_writers.utils")
NXclassUtils_logger.setLevel(logging.DEBUG)

# Define Timestamp dataset names
TSdset = Literal["start_time", "end_time", "end_time_estimated"]


def create_attributes(nxs_obj: h5py.Group | h5py.Dataset, names: Tuple, values: Tuple):
    """
    Create or overwrite attributes with additional metadata information.

    If any of the values are strings, these will first be converted into fixed-width \
    bytesrings.

    Args:
        nxs_obj (h5py.Group | h5py.Dataset): NeXus object to which the \
            attributes should be attached.
        names (Tuple): The names of the new attributes.
        values (Tuple): The attribute values asociated to the names.
    """
    for n, v in zip(names, values):
        if isinstance(v, str):
            # If a string, convert to a numpy bytestring
            v = np.bytes_(v)
        h5py.AttributeManager.create(nxs_obj, name=n, data=v)


def set_dependency(dep_info: str, path: str = None) -> np.bytes_:
    """
    Define value for "depends_on" attribute.
    If the attribute points to the head of the dependency chain, simply pass \
        "." for dep_info.

    Args:
        dep_info (str): The name of the transformation upon which the current \
            one depends on.
        path (str): Where the transformation is. Set to None, if passed it \
            points to location in the NeXus tree.
    Returns:
        The value to be passed to the attribute "depends_on" as a fixed-width \
            bytestring.
    """
    if dep_info == ".":
        return np.bytes_(".")
    if path:
        if path.endswith("/") is False:
            path += "/"
        return np.bytes_(path + dep_info)
    else:
        return np.bytes_(dep_info)


def calculate_origin(
    beam_center_fs: List | Tuple,
    fs_pixel_size: List | Tuple,
    fast_axis_vector: Tuple,
    slow_axis_vector: Tuple,
    mode: str = "1",
) -> Tuple[List, float]:
    """
    Calculate the offset of the detector.

    This function returns the detector origin array, which is saved as the \
        vector attribute of the module_offset field.
    The value to set the module_offset to is also returned: the magnitude of \
        the displacement if the vector is normalized, 1.0 otherwise
    Assumes that fast and slow axis vectors have already been converted to \
        mcstas if needed.

    Args:
        beam_center_fs (List | Tuple): Beam center position in fast and slow direction.
        fs_pixel_size (List | Tuple): Pixel size in fast and slow direction, in m.
        fast_axis_vector (Tuple): Fast axis vector.
        slow_axis_vector (Tuple): Slow axis vector.
        mode (str, optional): Decide how origin should be calculated.
            If set to "1" the displacement vector is un-normalized \
                and the offset value set to 1.0.
            If set to "2" the displacement is normalized and the \
                offset value is set to the magnitude of the displacement.
            Defaults to "1".

    Returns:
        det_origin (List): Displacement of beam center, vector attribute of module_offset.
        offset_val (float): Value to assign to module_offset, depending whether \
            det_origin is normalized or not.
    """
    # what was calculate module_offset
    x_scaled = beam_center_fs[0] * fs_pixel_size[0]
    y_scaled = beam_center_fs[1] * fs_pixel_size[1]
    # Detector origin
    det_origin = x_scaled * np.array(fast_axis_vector) + y_scaled * np.array(
        slow_axis_vector
    )
    det_origin = list(-det_origin)
    if mode == "1":
        offset_val = 1.0
    else:
        offset_val = math.hypot(*det_origin[:-1])
    return det_origin, offset_val


def find_number_of_images(datafile_list: List[Path], entry_key: str = "data") -> int:
    """
    Calculate total number of images when there's more than one input HDF5 file.

    Args:
        datafile_list (List[Path]): List of paths to the input image files.
        entry_key (str):    Key for the location of the images inside the \
            data files. Defaults to "data".

    Returns:
        num_images (int): Total number of images.
    """
    num_images = 0
    for filename in datafile_list:
        with h5py.File(filename, "r") as f:
            num_images += f[entry_key].shape[0]
    return int(num_images)


def calculate_estimated_end_time(
    start_time: datetime | str, tot_collection_time: float
) -> str:
    time_format = r"%Y-%m-%dT%H:%M:%SZ"

    if isinstance(start_time, str):
        start_time = start_time.format("%Y-%m-%dT%H:%M:%S")
        start_time = datetime.strptime(start_time.strip("Z"), time_format.strip("Z"))

    est_end = start_time + timedelta(seconds=tot_collection_time)
    return est_end.strftime(time_format)


def mask_and_flatfield_writer(
    nxdet_grp: h5py.Group,
    dset_name: str,
    dset_data: str | ArrayLike,
    applied_val: bool,
):
    """ Utility function to write mask or flatfield to NXdetector group for \
        image data when not already linked to the _meta.h5 file.
    If the pixel_mask/flatfield data is passed as a string, it will be assumed \
        to be a file path and the writer will try to set up an external link to it.

    Args:
        nxdet_grp (h5py.Group): Handle to HDF5 NXdetector group.
        dset_name (str): Name of the new field/dataset to be written.
        dset_data (str | ArrayLike): Dataset data to be written in the field. \
            Can be a string or an array-like dataset. \
            If the data type is a numpy ndarray, it will be compressed before writing.
        applied_val (bool): Value to write to `{flatfield,pixel_mask}_applied` fields.
    """
    if dset_data is None:
        NXclassUtils_logger.warning(
            f"""
            No copy of the {dset_name} has been found, either as a file or dataset.
            Fields {dset_name} and {dset_name}_applied will not be written to file.
            """
        )
        return
    nxdet_grp.create_dataset(
        f"{dset_name}_applied",
        data=applied_val,
    )
    NXclassUtils_logger.debug(f"{dset_name}_applied set to: {applied_val}.")
    if isinstance(dset_data, str):
        try:
            link_path = Path(dset_data)
            NXclassUtils_logger.debug(
                f"Setting external link for {dset_name} to {link_path}."
            )
            nxdet_grp[dset_name] = h5py.ExternalLink(link_path.name, "/")
        except Exception as e:
            NXclassUtils_logger.error(
                f"Impossible to write external link to {dset_data} for {dset_name}."
                "Field {dset_name} not written."
            )
            NXclassUtils_logger.error(f"{e}", exc_info=1)
    elif isinstance(dset_data, np.ndarray):
        NXclassUtils_logger.debug(f"Writing a compressed copy of array in {dset_name}.")
        write_compressed_copy(nxdet_grp, dset_name, data=dset_data)
    else:
        NXclassUtils_logger.debug(
            f"{dset_name} of type {type(dset_data)}, writing as is."
        )
        nxdet_grp.create_dataset(dset_name, data=dset_data)
    return


def mask_and_flatfield_writer_for_event_data(
    nxdet_grp: h5py.Group,
    dset_name: str,
    dset_data_file: str,
    applied_val: bool,
    wdir: Path,
    detector_name: str = "tristan",
):
    if dset_data_file is None:
        NXclassUtils_logger.warning(
            f"No {dset_name} data file passed; {dset_name} won't be written."
        )
        return

    nxdet_grp.create_dataset(f"{dset_name}_applied", data=applied_val)
    NXclassUtils_logger.debug(
        f"Looking for file {dset_data_file} in {wdir.as_posix()}."
    )
    filename = [
        wdir / dset_data_file for f in wdir.iterdir() if dset_data_file == f.name
    ]
    if filename:
        NXclassUtils_logger.debug(f"File {dset_name} found in working directory.")
        write_compressed_copy(
            nxdet_grp,
            dset_name,
            filename=filename[0],
            filter_choice="blosc",
            dset_key="image",
        )
    else:
        NXclassUtils_logger.warning(
            f"No {dset_name} file found in working directory."
            "Writing an ExternalLink."
        )
        file_loc = Path(dset_data_file)
        image_key = "image" if "tristan" in detector_name.lower() else "/"
        nxdet_grp[dset_name] = h5py.ExternalLink(file_loc.name, image_key)
    return


# Copy and compress a dataset inside a specified NXclass
def write_compressed_copy(
    nxgroup: h5py.Group,
    dset_name: str,
    data: ArrayLike = None,
    filename: Path | str = None,
    filter_choice: str = "bitshuffle",
    dset_key: str = "image",
    **kwargs,
):
    """
    Write a compressed copy of some dataset in the desired HDF5 group, using \
        the filter of choice with lz4 compression. Available filters at this \
        time include "Blosc" and "Bitshuffle" (default).
    The main application for this function in nexgen is to write a compressed \
        copy of a pixel mask or a flatfield file/dataset directly into the \
        NXdetector group of a NXmx NeXus file.
    The data and filename arguments are mutually exclusive as only one of them \
        can be used as input.
    If a filename is passed, it is also required to pass the key for the \
        relevant dataset to be copied. Failure to do so will result in nothing being \
        written to the NeXus file.

    Args:
        nxgroup (h5py.Group): Handle to HDF5 group.
        dset_name (str): Name of the new dataset to be written.
        data (ArrayLike, optional): Dataset to be compressed. Defaults to None.
        filename (Path | str, optional): Filename containing the dataset to be \
            compressed into the NeXus file. Defaults to None.
        filter_choice (str, optional): Filter to be used for compression. \
            Either blosc or bitshuffle. Defaults to bitshuffle.
        dset_key (str, optional): Dataset name inside the passed file. \
            Defaults to "image".

    Keyword Args:
        block_size (int, optional): Number of elements per block, it needs to \
            be divisible by 8. Needed for Bitshuffle filter. Defaults to 0.

    Raises:
        ValueError: If both a dataset and a filename have been passed to the function.
    """
    if data is not None and filename is not None:
        raise ValueError(
            "The dset and filename arguments are mutually exclusive."
            "Please pass only the one from which the data should be copied."
        )

    if filename and not dset_key:
        NXclassUtils_logger.warning(
            f"Missing key to find the dataset to be copied inside {filename}. {dset_name} will not be written into the NeXus file."
        )
        return

    if filename:
        with h5py.File(filename, "r") as fh:
            data = fh[dset_key][()]

    if filter_choice.lower() == "blosc":
        nxgroup.create_dataset(
            dset_name, data=data, **Blosc(cname="lz4", shuffle=Blosc.BITSHUFFLE)
        )
    elif filter_choice.lower() == "bitshuffle":
        block_size = (
            0 if "block_size" not in list(kwargs.keys()) else kwargs["block_size"]
        )
        nxgroup.create_dataset(
            dset_name, data=data, **Bitshuffle(nelems=block_size, cname="lz4")
        )
    else:
        NXclassUtils_logger.warning(
            "Unknown filter choice, no dataset will be written."
        )
        return
    NXclassUtils_logger.info(
        f"A compressed copy of the {dset_name} has been written into the NeXus file."
    )


def add_sample_axis_groups(nxsample: h5py.Group, axis_list: List[Axis]):
    """
    Add non-standard "sample_{phi,omega,...}" groups to NXsample. These may be needed for \
    some autoprocessing tools to work correctly.

    Args:
        nxsample (h5py.Group): NeXus NXsample group.
        axis_list (List[Axis]): List of goniometer axes.
    """
    NXclassUtils_logger.debug("Add non-standard fields for autoPROC to work.")
    nxtransf = nxsample["transformations"]
    for ax in axis_list:
        grp_name = f"sample_{ax.name[-1]}" if "sam" in ax.name else f"sample_{ax.name}"
        nx_ax = nxsample.require_group(grp_name)
        # NOTE: NX_class here set to NXtransformations instead of NXpositioner
        # One step closer to standard.  TO BE TESTED
        create_attributes(nx_ax, ("NX_class",), ("NXtransformations",))
        nx_ax[ax.name] = nxtransf[ax.name]
        if f"{ax.name}_end" in nxtransf.keys():
            nx_ax[f"{ax.name}_end"] = nxtransf[f"{ax.name}_end"]
        if f"{ax.name}_increment_set" in nxtransf.keys():
            nx_ax[f"{ax.name}_increment_set"] = nxtransf[f"{ax.name}_increment_set"]
